# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import bz2 as bzp2
import datetime as dttm
import shelve as shlv
import sys as sstm
from pathlib import Path as path_t
from typing import Any, Protocol, Union

from cell_tracking_BC.in_out.text.logger import LOGGER


class jsonable_t(Protocol):
    def AsJsonString(self) -> str:
        ...

    def NewFromJsonString(self, jsoned: str, /) -> jsonable_t:
        ...


SHOULD_JSON = False
_PIPELINE_STAGE = 0


def CacheDictionaryForMain(
    main: Union[str, path_t],
    /,
    *,
    folder_name: str = "_runtime",
    folder_postfix: str = "",
    cache_postfix: str = "",
) -> shlv.Shelf:
    """
    Call with main equal to __file__
    """
    if isinstance(main, str):
        main = path_t(main)

    cache_folder = main.parent / (folder_name + folder_postfix)
    if not cache_folder.exists():
        cache_folder.mkdir()
    elif not cache_folder.is_dir():
        print(f"{cache_folder}: Not a suitable cache folder")
        sstm.exit(-1)

    return shlv.open(str(cache_folder / (main.stem + cache_postfix)))


def CloseCache(cache: Union[shlv.Shelf, Any]) -> None:
    """"""
    if isinstance(cache, shlv.Shelf):
        cache.close()


def Save(element: jsonable_t, title: str, folder: path_t, /) -> None:
    """"""
    global _PIPELINE_STAGE

    if SHOULD_JSON:
        _PIPELINE_STAGE += 1
        jsoned = element.AsJsonString()
        compressed = bzp2.compress(jsoned.encode())
        with open(folder / f"{_PIPELINE_STAGE}-{title}.json", "bw") as accessor:
            accessor.write(compressed)

    # import sys as sstm
    # from cell_tracking_BC.test.test_jsonization import CompareInstances
    # decoded = type(element).NewFromJsonString(jsoned)
    #
    # n_issues = CompareInstances(element, decoded)
    # if n_issues > 0:
    #     sstm.exit(1)


def Load(type_: jsonable_t, folder: path_t, /) -> object:
    """"""
    global _PIPELINE_STAGE

    _PIPELINE_STAGE += 1
    where = tuple(folder.glob(f"{_PIPELINE_STAGE}-*.json"))
    if where.__len__() != 1:
        raise RuntimeError(
            f"{folder}: Folder should contain a unique stage {_PIPELINE_STAGE}."
        )
    where = where[0]

    with open(folder / where, "br") as accessor:
        compressed = accessor.read()

    return type_.NewFromJsonString(bzp2.decompress(compressed).decode())


def ReplayFolderForMain(
    main: Union[str, path_t],
    folder_name: str,
    /,
    *,
    base_folder_name: str = "_replay",
    base_folder_postfix: str = "",
    play_date: str = None,
) -> path_t:
    """
    Call with main equal to __file__
    play_date: Can be None (play mode), or (replay mode) "latest" or a date-time in iso format with
        timespec='milliseconds'
    """
    if isinstance(main, str):
        main = path_t(main)

    output = main.parent / (base_folder_name + base_folder_postfix)
    if not output.exists():
        output.mkdir()
    elif not output.is_dir():
        print(f"{output}: Not a suitable replay folder")
        sstm.exit(-1)

    output /= folder_name
    if not output.exists():
        output.mkdir()
    elif not output.is_dir():
        print(f"{output}: Not a suitable replay folder")
        sstm.exit(-1)

    if play_date is None:
        output /= dttm.datetime.now().isoformat(timespec="milliseconds")
        if not output.exists():
            output.mkdir()
        elif not output.is_dir():
            print(f"{output}: Not a suitable replay folder")
            sstm.exit(-1)
        LOGGER.info(f"SAVING PIPELINE STAGES in subfolder {output.name}")
    elif play_date == "latest":
        everything = sorted(output.glob("*"))
        output /= everything[-1]
    else:
        output /= play_date
    if play_date is not None:
        if not output.is_dir():
            print(f"{output}: Not a valid replay folder")
            sstm.exit(-1)
        LOGGER.warn(f"REPLAYING PIPELINE from {output.name}")

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numbers as nmbr
import tempfile as temp
from pathlib import Path as path_t
from typing import Dict, Optional, Sequence, Union

import xlsxwriter as xlsx
from xlsxwriter.format import Format as xlsx_format_t
from xlsxwriter.workbook import Workbook as workbook_t
from xlsxwriter.worksheet import Worksheet as worksheet_t

from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.uid import AlphaColumnFromIndex
from cell_tracking_BC.type.cell import state_e
from cell_tracking_BC.type.sequence import sequence_t
from cell_tracking_BC.type.track.single import single_track_t


event_response_h = Dict[int, Optional[Sequence[float]]]
division_times_h = Dict[int, Sequence[int]]
death_time_h = Dict[int, int]


_INVALID_TRACK = "Invalid Track"
_PRUNED_TRACK = "Pruned Track"

_DASH_TYPES = (
    "solid",
    "round_dot",
    "square_dot",
    "dash",
    "dash_dot",
    "long_dash",
    "long_dash_dot",
    "long_dash_dot_dot",
)
_N_DASH_TYPES = _DASH_TYPES.__len__()


def SaveCellFeaturesToXLSX(
    path: Union[str, path_t],
    sequence: sequence_t,
    /,
    *,
    feature: Union[str, Sequence[str]] = None,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if feature is None:
        features = sequence.available_cell_features
    elif isinstance(feature, str):
        features = (feature,)
    else:
        features = feature

    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    workbook = xlsx.Workbook(str(path))

    pruned_format = workbook.add_format({"bg_color": "gray"})
    division_format = workbook.add_format({"bg_color": "blue"})
    death_format = workbook.add_format({"bg_color": "red"})

    for feature in features:
        _SaveCellFeature(feature, sequence, workbook, pruned_format, division_format, death_format)

    workbook.close()


def _SaveCellFeature(
    feature: str,
    sequence: sequence_t,
    workbook: workbook_t,
    pruned_format: Optional[xlsx_format_t],
    division_format: Optional[xlsx_format_t],
    death_format: Optional[xlsx_format_t],
    /,
) -> None:
    """"""
    evolutions = sequence.CellFeature(feature, geometric_mode = True)
    if evolutions.__len__() == 0:
        LOGGER.warn(f"{feature}: Feature without associated tracks.")
        return

    first_label = tuple(evolutions.keys())[0]
    first_feature = evolutions[first_label][1][0]
    if not isinstance(first_feature, nmbr.Number):
        feature_type = type(first_feature).__name__
        LOGGER.warn(f'Feature of type "{feature_type}": Unhandled')
        return

    sheet_name = _SheetNameFromLongName(feature)
    worksheet = workbook.add_worksheet(sheet_name)

    max_label = max(evolutions.keys())
    labels = sequence.tracks.labels
    per_row_limits = {}
    for label in range(1, max_label + 1):
        if label not in evolutions:
            worksheet.write_string(label - 1, 0, _INVALID_TRACK)
            continue
        if label not in labels:
            worksheet.write_string(label - 1, 0, _PRUNED_TRACK)
            continue

        track, evolution = evolutions[label]
        worksheet.write_row(label - 1, track._geometric_root_time_point, evolution)

        _SetStateBasedCellFormat(
            worksheet,
            label - 1,
            track,
            evolution,
            pruned_format,
            division_format,
            death_format,
        )

        per_row_limits[label] = (
            track._geometric_root_time_point,
            track._geometric_root_time_point + evolution.__len__() - 1,
        )
    next_available_row = max_label

    worksheet.write_string(next_available_row, 0, "END")
    next_available_row += 2  # With margin

    _AddCellStateLegend(worksheet, next_available_row, pruned_format, division_format, death_format)
    next_available_row += 3  # With margin

    if per_row_limits.__len__() > 0:
        chart = workbook.add_chart({"type": "line"})
        for l_idx, (row, (min_col, max_col)) in enumerate(per_row_limits.items()):
            min_col = AlphaColumnFromIndex(min_col)
            max_col = AlphaColumnFromIndex(max_col)
            chart.add_series(
                {
                    "name": str(row),
                    "values": f"='{sheet_name}'!${min_col}${row}:${max_col}${row}",
                    "line": {
                        "width": 1.0,
                        "dash_type": _DASH_TYPES[l_idx % _N_DASH_TYPES],
                    },
                }
            )
        worksheet.insert_chart(f"A{next_available_row}", chart)


def SaveCellEventsToXLSX(
    path: Union[str, path_t],
    sequence: sequence_t,
    /,
    *,
    division_response: str = None,
    death_response: str = None,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    workbook = xlsx.Workbook(str(path))
    pruned_format = workbook.add_format({"bg_color": "gray"})
    division_format = workbook.add_format({"bg_color": "blue"})
    death_format = workbook.add_format({"bg_color": "red"})

    division_time_points = sequence.tracks.DivisionTimePoints()
    death_time_points = sequence.tracks.DeathTimePoints(
        sequence.length, with_living_leaves=True
    )
    for SaveEvents, time_point in zip(
        (_SaveDivisionEvents, _SaveDeathEvents),
        (division_time_points, death_time_points),
    ):
        SaveEvents(workbook, time_point)

    n_divisions = sequence.tracks.n_dividing_cells
    _SaveEventCounts(workbook, n_divisions, death_time_points)

    for event, response in zip(
        ("division", "death"), (division_response, death_response)
    ):
        if response is not None:
            _SaveEventResponse(
                event,
                workbook,
                response,
                sequence,
                pruned_format,
                division_format,
                death_format,
            )

    workbook.close()


def _SaveDivisionEvents(
    workbook: xlsx.Workbook,
    division_time_points: division_times_h,
    /,
) -> None:
    """"""
    if division_time_points.__len__() == 0:
        LOGGER.warn("Division Events: No associated tracks.")
        return

    sheet_name = _SheetNameFromLongName("division times")
    worksheet = workbook.add_worksheet(sheet_name)

    max_label = max(division_time_points.keys())
    for label in range(1, max_label + 1):
        if label not in division_time_points:
            worksheet.write_string(label - 1, 0, f"{_INVALID_TRACK} or {_PRUNED_TRACK}")
            continue

        time_points = division_time_points[label]
        if (time_points is None) or (time_points.__len__() == 0):
            worksheet.write_string(label - 1, 0, "No Divisions")
        else:
            worksheet.write_row(label - 1, 0, time_points)

    worksheet.write_string(max_label, 0, "END")


def _SaveDeathEvents(
    workbook: xlsx.Workbook, death_time_points: death_time_h, /
) -> None:
    """"""
    if death_time_points.__len__() == 0:
        LOGGER.warn("Death Events: No associated tracks.")
        return

    sheet_name = _SheetNameFromLongName("death time")
    worksheet = workbook.add_worksheet(sheet_name)

    max_label = max(death_time_points.keys())
    for label in range(1, max_label + 1):
        if label not in death_time_points:
            worksheet.write_string(label - 1, 0, f"{_INVALID_TRACK} or {_PRUNED_TRACK}")
            continue

        time_point = death_time_points[label]
        if time_point is None:
            worksheet.write_string(label - 1, 0, "No Death")
        else:
            worksheet.write_number(label - 1, 0, time_point)

    worksheet.write_string(max_label, 0, "END")


def _SaveEventCounts(
    workbook: xlsx.Workbook,
    n_divisions: int,
    death_time_points: death_time_h,
    /,
) -> None:
    """"""
    sheet_name = _SheetNameFromLongName("event counts")
    worksheet = workbook.add_worksheet(sheet_name)

    n_deaths_pattern = 0
    n_deaths_track = 0
    for time_point in death_time_points.values():
        if time_point is not None:
            if time_point >= 0:
                n_deaths_pattern += 1
            else:
                n_deaths_track += 1

    for r_idx, (title, value) in enumerate(
        zip(
            ("divisions", "death (pattern)", "death (geometric)", "death"),
            (
                n_divisions,
                n_deaths_pattern,
                n_deaths_track,
                n_deaths_pattern + n_deaths_track,
            ),
        )
    ):
        worksheet.write_string(r_idx, 0, title)
        worksheet.write_number(r_idx, 1, value)


def _SaveEventResponse(
    event: str,
    workbook: xlsx.Workbook,
    name: str,
    sequence: sequence_t,
    pruned_format: xlsx_format_t,
    division_format: xlsx_format_t,
    death_format: xlsx_format_t,
    /,
) -> None:
    """"""
    sheet_name = _SheetNameFromLongName(f"{event} response")
    worksheet = workbook.add_worksheet(sheet_name)

    responses = sequence.TrackFeature(name)
    if responses.__len__() == 0:
        LOGGER.warn(f"{event.title()} Event Responses: No associated tracks.")
        return

    max_label = max(responses.keys())
    labels = sequence.tracks.labels
    per_row_limits = {}
    for label in range(1, max_label + 1):
        if label not in responses:
            worksheet.write_string(label - 1, 0, _INVALID_TRACK)
            continue
        if label not in labels:
            worksheet.write_string(label - 1, 0, _PRUNED_TRACK)
            continue

        response = responses[label]
        if response is None:
            worksheet.write_string(label - 1, 0, "Track too Short for Valid Response")
        else:
            track = sequence.tracks.TrackWithLabel(label, geometric_mode = True)
            worksheet.write_row(label - 1, track._geometric_root_time_point, response)

            _SetStateBasedCellFormat(
                worksheet,
                label - 1,
                track,
                response,
                pruned_format,
                division_format,
                death_format,
            )

            per_row_limits[label] = (
                track._geometric_root_time_point,
                track._geometric_root_time_point + response.__len__() - 1,
            )
    next_available_row = max_label

    worksheet.write_string(next_available_row, 0, "END")
    next_available_row += 2  # With margin

    _AddCellStateLegend(worksheet, next_available_row, pruned_format, division_format, death_format)
    next_available_row += 3  # With margin

    if per_row_limits.__len__() > 0:
        chart = workbook.add_chart({"type": "line"})
        for l_idx, (row, (min_col, max_col)) in enumerate(per_row_limits.items()):
            max_col = AlphaColumnFromIndex(max_col)
            chart.add_series(
                {
                    "name": str(row),
                    "values": f"='{sheet_name}'!${min_col}${row}:${max_col}${row}",
                    "line": {
                        "width": 1.0,
                        "dash_type": _DASH_TYPES[l_idx % _N_DASH_TYPES],
                    },
                }
            )
        worksheet.insert_chart(f"A{next_available_row}", chart)


def _SheetNameFromLongName(name: str, /) -> str:
    """
    Sheet names cannot exceed 31 characters in length:
    xlsxwriter.exceptions.InvalidWorksheetName: Excel worksheet name 'cfp-background_over_yfp-background' must be <= 31 chars.
    """
    LIMIT = 31

    if name.__len__() <= LIMIT:
        return name

    return f"{name[:(LIMIT-3)]}..."


def _SetStateBasedCellFormat(
    worksheet: worksheet_t,
    row: int,
    track: single_track_t,
    values: Sequence,
    pruned_format: xlsx_format_t,
    division_format: xlsx_format_t,
    death_format: xlsx_format_t,
    /,
) -> None:
    """"""
    root_time_point = track._geometric_root_time_point
    for time_point, cell in enumerate(track, start=root_time_point):
        if cell.state is state_e.pruned:
            cell_format = pruned_format
        elif cell.state is state_e.dividing:
            cell_format = division_format
        elif cell.state is state_e.dead:
            cell_format = death_format
        else:
            cell_format = None

        if cell_format is not None:
            value = values[time_point - root_time_point]
            worksheet.write_number(row, time_point, value, cell_format)


def _AddCellStateLegend(
    worksheet: worksheet_t,
    row: int,
    pruned_format: Optional[xlsx_format_t],
    division_format: Optional[xlsx_format_t],
    death_format: Optional[xlsx_format_t],
    /,
) -> None:
    """"""
    for col, state, cell_format in zip(
        range(3),
        ("Dividing", "Dead", "Pruned"),
        (division_format, death_format, pruned_format),
    ):
        if cell_format is not None:
            worksheet.write_string(row, col, state, cell_format)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from pathlib import Path as path_t
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as nmpy
import skimage.measure as msre
import skimage.morphology as mrph

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.uid import Identity
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.compartment import _AdjustMap
from cell_tracking_BC.type.segmentation import segmentation_t


array_t = nmpy.ndarray
inner_outer_associations_h = Dict[int, Union[int, Tuple[int, int]]]


class frame_t(array_t):
    """
    A Frame is one plane, or channel, of a (potentially multi-channel) image
    Attributes: path (only str or possibly path_t?), cells (list[cell_t]), runtime (dict[str, Any])
    """

    path: Optional[path_t]
    cells = List[cell_t]
    runtime = Dict[str, Any]

    def __new__(cls, array: array_t, /) -> frame_t:
        """"""
        if (n_dims := array.ndim) != 2:
            raise ValueError(
                f"{n_dims}: Invalid number of dimensions of frame with shape {array.shape}. "
                f"Expected=2=ROWSxCOLUMNS."
            )

        instance = nmpy.asarray(array).view(cls)
        instance.path = None
        instance.cells = []
        instance.runtime = {}

        return instance

    def __array_finalize__(self, array: Optional[array_t], /) -> None:
        """"""
        if array is None:
            return
        self.path = getattr(array, "path", None)
        self.cells = getattr(array, "cells", [])  # Used to be None. Why?
        self.runtime = getattr(array, "runtime", {})  # Used to be missing. Why?

    @classmethod
    def NewFromJsonString(cls, jsoned: str, /) -> frame_t:
        """"""
        return jsnr.ObjectFromJsonString(
            jsoned,
            builders={
                cls.__name__: cls.NewFromJsonDescription,
                "cell_t": cell_t.NewFromJsonString,
            },
        )

    @classmethod
    def NewFromJsonDescription(cls, description: dict, /) -> frame_t:
        """"""
        instance = cls(description["frame"])
        instance.path = description["path"]
        instance.cells = description["cells"]
        instance.runtime = description["runtime"]

        return instance

    def AsJsonString(self) -> str:
        """"""
        as_dict = {
            "frame": nmpy.array(self),
            "path": self.path,
            "cells": self.cells,
            "runtime": self.runtime,
        }

        return jsnr.JsonStringOf(as_dict, true_type=self)

    def ApplyTransform(self, Transform: transform_h, /, **kwargs) -> None:
        """"""
        transformed = Transform(self, **kwargs)
        self[...] = transformed[...]

    def AddCellsFromSegmentation(
        self,
        segmentation: segmentation_t,
    ) -> None:
        """
        Segmentation are supposed to be binary (as opposed to already labeled)
        """
        # --- Generic inner and outer segmentation list for all options
        non_nones, parameters = segmentation.NonNoneAsList()

        # --- Segmentation labeling and (if needed) inner-outer label associations
        labeleds = []
        n_objects_per_sgm = []
        for non_none in non_nones:
            labeled, n_objects = mrph.label(non_none, return_num=True, connectivity=1)
            labeleds.append(labeled)
            n_objects_per_sgm.append(n_objects)
        if labeleds.__len__() > 1:
            inner_label_of_outer = _InnerOuterAssociations(labeleds, n_objects_per_sgm)
            _CheckInnerOuterAssociations(inner_label_of_outer, n_objects_per_sgm)
        else:
            inner_label_of_outer = None

        # --- Cell creation and addition
        for outer_label in range(1, n_objects_per_sgm[-1] + 1):
            if inner_label_of_outer is None:
                labels = (outer_label,)
            else:
                labels = (inner_label_of_outer[outer_label], outer_label)
            kwargs = {}
            additional_nucleus = None
            for parameter, labeled, label in zip(parameters, labeleds, labels):
                label: Union[int, Tuple[int, int]]
                if isinstance(label, int):
                    kwargs[parameter] = labeled == label
                else:
                    kwargs[parameter] = labeled == label[0]
                    additional_nucleus = label[1]
            cell = cell_t.NewFromMaps(labels[-1], **kwargs)
            if additional_nucleus is not None:
                cell.AddNucleus(labeleds[0] == additional_nucleus)
            self.cells.append(cell)

    def Description(self, /, *, margins: Sequence[float] = None) -> Dict[str, Any]:
        """"""
        output = {
            "cell_labels": tuple((_cll.label, _cll.centroid) for _cll in self.cells)
        }

        contours = {}
        if margins is None:
            margins = (0,)
        for margin in margins:
            for_margin = []
            for cell in self.cells:
                if (margin is None) or (margin <= 0.0):
                    bb_map = nmpy.pad(cell.BBMap().astype(nmpy.int8), 1)
                    if (margin is not None) and (margin < 0.0):
                        bb_map = _AdjustMap(bb_map, cell.area, margin)
                    around_cell = _LargestContourOfMap(bb_map)
                    around_cell += (
                        cell.bb_slices[0].start - 1,
                        cell.bb_slices[1].start - 1,
                    )
                else:
                    cell_map = cell.Map(self.shape, as_boolean=False, margin=margin)
                    around_cell = _LargestContourOfMap(cell_map)
                for_margin.append(around_cell)
            contours[margin] = tuple(for_margin)
            # contour_map = nmpy.zeros(self.shape, dtype = nmpy.int8)
            # for cell in self.cells:
            #     contour_map[cell.Map(self.shape, as_boolean = True, margin = margin)] = 1
            # eroded = mrph.binary_erosion(contour_map)  # .astype(nmpy.int8, copy=False)
            # contour_map[eroded] = 0
            # contours[margin] = nmpy.nonzero(contour_map)  # msre.find_contours(segmentation, level=0.5)
        output["cell_contours"] = contours

        return output

    def __eq__(self, other) -> bool:
        """"""
        if hasattr(other, "AsJsonString"):
            return self.AsJsonString() == other.AsJsonString()

        return False

    def __repr__(self) -> str:
        """"""
        return Identity(self)

    def __str__(self) -> str:
        """"""
        return (
            f"{self.__repr__()}:\n"
            f"    {self.path=}\n"
            f"    {self.shape=}\n"
            f"    {self.cells.__len__()=}"
        )


transform_h = Callable[[frame_t, Dict[str, Any]], array_t]


def _InnerOuterAssociations(
    segmentations: List[array_t], n_objects_per_sgm: Sequence[int], /
) -> inner_outer_associations_h:
    """"""
    output = {}

    for inner_label in range(1, n_objects_per_sgm[0] + 1):
        dilated = mrph.dilation(segmentations[0] == inner_label)
        outer_label = nmpy.amax(segmentations[1][dilated])
        if outer_label in output:
            output[outer_label] = (
                output[outer_label],
                inner_label,
            )
            if output[outer_label].__len__() > 2:
                raise RuntimeError("Cell with more than 2 nuclei")
        else:
            output[outer_label] = inner_label

    return output


def _CheckInnerOuterAssociations(
    inner_label_of_outer: inner_outer_associations_h,
    n_objects_per_sgm: Sequence[int],
    /,
) -> None:
    """"""
    if n_objects_per_sgm[0] != n_objects_per_sgm[1]:
        LOGGER.warn(
            f"Mismatch in number of segmented objects: {n_objects_per_sgm}; "
            f"Might be due to cell divisions"
        )

    inner_labels = []
    for label in inner_label_of_outer.values():
        if isinstance(label, int):
            inner_labels.append(label)
        else:
            inner_labels.extend(label)
    outer_labels = tuple(inner_label_of_outer.keys())
    n_inner_labels = nmpy.unique(inner_labels).size
    n_outer_labels = nmpy.unique(outer_labels).size
    if (n_missing_associations := n_objects_per_sgm[0] - n_inner_labels) > 0:
        LOGGER.warn(
            f"{n_missing_associations} inner comportment(s) not associated with outer compartments"
        )
    if (n_missing_associations := n_objects_per_sgm[1] - n_outer_labels) > 0:
        LOGGER.warn(
            f"{n_missing_associations} outer comportment(s) without inner compartment(s)"
        )


def _LargestContourOfMap(map_: array_t, /) -> array_t:
    """"""
    contours = msre.find_contours(map_, level=0.5)
    if (n_contours := contours.__len__()) > 1:
        lengths = tuple(_ctr.shape[0] for _ctr in contours)
        largest = lengths.index(max(lengths))

        return contours[largest]
    elif n_contours > 0:
        return contours[0]

    raise ValueError("Cell map has no 0.5-isocontour")



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import itertools as ittl

import matplotlib.pyplot as pypl
import numpy as nmpy

import cell_tracking_BC.type.cell as cll_
from cell_tracking_BC.type.cell import cell_t


array_t = nmpy.ndarray


shape = (100, 100)

nucleus_map = nmpy.zeros(shape, dtype=nmpy.bool_)
nucleus_map[20:28, 35:47] = True
nucleus_map[50:58, 60:72] = True

cytoplasm_map = nmpy.zeros(shape, dtype=nmpy.bool_)
cytoplasm_map[15:35, 29:55] = True
filled_cytoplasm_map = cytoplasm_map.copy()
cytoplasm_map[45:65, 54:80] = True
cytoplasm_map[nucleus_map] = False

cell_map = cytoplasm_map.copy()
cell_map[nucleus_map] = True

empty_map = nmpy.zeros_like(cell_map)

_, all_axes = pypl.subplots(ncols=3)
for axes, map_ in zip(all_axes, (cell_map, cytoplasm_map, nucleus_map)):
    axes.matshow(map_)

for maps in ittl.product((cell_map, None), (cytoplasm_map, None), (nucleus_map, None)):
    try:
        cell, cytoplasm, nucleus = maps
        validateds = cll_.CellAndCytoplasmMapsFromCombinations(
            cell=cell,
            cytoplasm=cytoplasm,
            nucleus=nucleus,
            should_return_nucleus=True,
        )
        _, all_axes = pypl.subplots(ncols=3)
        for axes, map_, validated in zip(all_axes, maps, validateds):
            if validated is None:
                axes.matshow(empty_map)
            else:
                axes.matshow(validated)
            axes.set_title(type(map_).__name__[:4])
    except ValueError as exc:
        types = tuple(type(_elm).__name__[:4] for _elm in maps)
        print(f"{types} => {exc}")

try:
    print("Filled cytoplasm map: Should raise an exception...")
    _ = cell_t.NewFromMaps(1, cytoplasm_map = filled_cytoplasm_map)
except ValueError as exc:
    print(f"    Exception: {exc}")

pypl.show()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import itertools as ittl
from operator import attrgetter as GetAttribute
from operator import itemgetter as ItemAt
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.standard.uid import Identity
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.track.all import any_track_h, structured_track_h
from cell_tracking_BC.type.track.forking import forking_track_t
from cell_tracking_BC.type.track.single import single_track_t
from cell_tracking_BC.type.track.unstructured import unstructured_track_t
from cell_tracking_BC.type.track.structured import (
    cells_w_optional_time_points_h,
    per_single_track_cell_h,
    per_single_track_cells_h,
)
from cell_tracking_BC.type.tracks.unstructured import unstructured_tracks_t


TrackIssues_h = Callable[
    [structured_track_h, dict], Optional[Union[str, Sequence[str]]]
]


@dtcl.dataclass(repr=False, eq=False)
class tracks_t(List[structured_track_h]):

    invalids: List[Tuple[any_track_h, Sequence[str]]] = dtcl.field(
        init=False, default_factory=list
    )
    fully_pruned: List[structured_track_h] = dtcl.field(
        init=False, default_factory=list
    )

    @classmethod
    def NewFromUnstructuredTracks(cls, tracks: unstructured_tracks_t, /) -> tracks_t:
        """"""
        instance = cls()

        next_single_track_label = 1
        for unstructured in tracks.track_iterator:
            issues = unstructured.Issues()
            if issues is None:
                (
                    forking_track,
                    next_single_track_label,
                ) = forking_track_t.NewFromUnstructuredTrack(
                    unstructured, next_single_track_label
                )

                if forking_track.n_leaves > 1:
                    instance.append(forking_track)
                else:
                    single_track = forking_track.AsSingleTrack()
                    instance.append(single_track)
            else:
                # unstructured.graph["issues"] = issues  # Old, per-track issue storage
                instance.invalids.append((unstructured, issues))

        return instance

    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> tracks_t:
        """"""
        return jsnr.ObjectFromJsonString(
            jsoned,
            builders={
                cls.__name__: cls.NewFromJsonDescription,
                "unstructured_track_t": unstructured_track_t.NewFromJsonString,
                "single_track_t": single_track_t.NewFromJsonString,
                "forking_track_t": forking_track_t.NewFromJsonString,
            },
        )

    @classmethod
    def NewFromJsonDescription(
        cls,
        description,
        /,
    ) -> tracks_t:
        """"""
        tracks, attributes = description

        instance = cls.__new__(cls)  # Do not use cls(tracks)
        instance.extend(tracks)
        for key, value in attributes.items():
            setattr(instance, key, value)

        return instance

    @property
    def labels(self) -> Sequence[int]:
        """"""
        output = []

        for track in self:
            output.extend(track.labels)

        return sorted(output)

    def RootCells(
        self, /, *, with_time_point: bool = False
    ) -> cells_w_optional_time_points_h:
        """"""
        if with_time_point:
            what = GetAttribute("root", "root_time_point")
            # output = map(GetAttribute("root", "root_time_point"), self)
            # output = ((_tck.root, _tck.root_time_point) for _tck in self)
        else:
            what = GetAttribute("root")
            # output = map(GetAttribute("root"), self)
            # output = (_tck.root for _tck in self)

        return tuple(map(what, self))

    def MarkDividingCells(
        self,
        before_deaths: bool,
        /,
        *,
        division_responses: Dict[int, Optional[Sequence[float]]] = None,
        lower_bound: float = None,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """"""
        for track in self:
            if isinstance(track, forking_track_t):
                track.MarkDividingCells(
                    before_deaths,
                    division_responses=division_responses,
                    lower_bound=lower_bound,
                    CombinedResponse=CombinedResponse,
                )

    @property
    def n_dividing_cells(self) -> int:
        """"""
        return sum(_tck.n_dividing_cells for _tck in self)

    def DividingCells(
        self,
        /,
        *,
        with_time_point: bool = False,
        per_single_track: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_single_track_cells_h]:
        """"""
        if per_single_track:
            output = {}
            AddDividing = output.update
        else:
            output = []
            AddDividing = output.extend

        for track in self:
            if isinstance(track, forking_track_t):
                dividing = track.DividingCells(
                    with_time_point=with_time_point, per_single_track=per_single_track
                )
                AddDividing(dividing)

        if with_time_point and not per_single_track:
            output.sort(key=ItemAt(1))

        return output

    def DivisionTimePoints(self) -> Dict[int, Optional[Sequence[int]]]:
        """"""
        output = {}

        for track in self:
            if isinstance(track, forking_track_t):
                time_points = track.DivisionTimePoints()
                output.update(time_points)
            else:
                output[track.label] = None

        return output

    def MarkDeadCells(
        self,
        death_responses: Dict[int, Optional[Sequence[float]]],
        lower_bound: float,
        after_divisions: bool,
        /,
        *,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """
        A death is set if the highest response at a cell among the adjacent single tracks is above the threshold.
        """
        t_idx = 0
        while t_idx < self.__len__():
            track = self[t_idx]
            track.MarkDeadCells(
                death_responses,
                lower_bound,
                after_divisions,
                CombinedResponse=CombinedResponse,
                called_from_educated_code=True,
            )
            if track.IsFullyPruned():
                self.fully_pruned.append(track)
                del self[t_idx]
            else:
                t_idx += 1

    def DeadCells(
        self,
        sequence_length: int,
        /,
        *,
        with_time_point: bool = False,
        with_living_leaves: bool = False,
        per_single_track: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_single_track_cell_h]:
        """"""
        if per_single_track:
            output = {}
            AddDead = output.update
        else:
            output = []
            AddDead = output.extend

        for track in self:
            dead = track.DeadCells(
                sequence_length,
                with_time_point=with_time_point,
                with_living_leaves=with_living_leaves,
                per_single_track=per_single_track,
            )
            AddDead(dead)

        return output

    def DeathTimePoints(
        self,
        sequence_length: int,
        /,
        *,
        with_living_leaves: bool = False,
    ) -> Dict[int, Optional[int]]:
        """"""
        output = {}

        for track in self:
            time_points = track.DeathTimePoints(
                sequence_length,
                with_living_leaves=with_living_leaves,
            )
            output.update(time_points)

        return output

    def LeafCells(
        self, /, *, with_time_point: bool = False
    ) -> Union[cells_w_optional_time_points_h, per_single_track_cells_h]:
        """"""
        leaves = []
        time_points = []
        for track in self:
            leaves.extend(track.leaves)
            time_points.extend(track.leaves_time_points)

        if with_time_point:
            return tuple(zip(leaves, time_points))

        return tuple(leaves)

    def TrackWithRoot(
        self, root: cell_t, /, *, tolerant_mode: bool = False
    ) -> Optional[structured_track_h]:
        """"""
        for track in self:
            if root is track.root:
                return track

        if tolerant_mode:
            return None

        raise ValueError(f"{root}: Not a root cell")

    def TrackWithCell(
        self, cell: cell_t, /, *, tolerant_mode: bool = False
    ) -> Optional[structured_track_h]:
        """"""
        for track in self:
            if cell in track:
                return track

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in any track")

    def TrackWithLeaf(
        self, leaf: cell_t, /, *, tolerant_mode: bool = False,
        geometric_mode: bool = False,
    ) -> Optional[single_track_t]:
        """
        Implicitly, it is SingleTrackWithLeaf
        """
        for track in self:
            if geometric_mode:
                leaves = track.geometric_leaves
            else:
                leaves = track.leaves
            if leaf in leaves:
                return track.TrackWithLeaf(leaf, check=False, geometric_mode=geometric_mode)
            # for cell in track.leaves:
            #     if leaf is cell:
            #         return track.TrackWithLeaf(cell, check=False)

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def TrackWithLabel(
        self, label: int, /, *, tolerant_mode: bool = False, geometric_mode: bool = False,
    ) -> Optional[single_track_t]:
        """
        Implicitly, it is SingleTrackWithLabel
        """
        for track in self:
            if isinstance(track, single_track_t):
                if track.label == label:
                    if geometric_mode:
                        return track
                    else:
                        return track.unpruned_piece
            else:
                if geometric_mode:
                    iterator = track.geometric_single_tracks_iterator
                else:
                    iterator = track.single_tracks_iterator
                for single_track in iterator:
                    if single_track.label == label:
                        return single_track

        if tolerant_mode:
            return None

        raise ValueError(f"{label}: Not a valid track label")

    @property
    def single_tracks_iterator(self) -> Iterator[single_track_t]:
        """"""
        for track in self:
            # Do not just yield track if it is a single track since it would not account for pruning
            for single_track in track.single_tracks_iterator:
                yield single_track

    @property
    def geometric_single_tracks_iterator(
        self,
    ) -> Iterator[single_track_t]:
        """"""
        for track in self:
            for single_track in track.geometric_single_tracks_iterator:
                yield single_track

    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """"""
        for track in self:
            if cell in track:
                return track.TrackLabelsContainingCell(
                    cell, check=False, tolerant_mode=tolerant_mode
                )

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Not a tracked cell")

    # def TrackLabelWithLeaf(
    #     self, leaf: cell_t, /, *, tolerant_mode: bool = False
    # ) -> Optional[int]:
    #     """
    #     TrackLabelWithLeaf: Implicitly, it is SingleTrackLabelWithLeaf
    #     """
    #     track = self.TrackWithLeaf(leaf, tolerant_mode=True)
    #     if track is not None:
    #         return track.label
    #
    #     if tolerant_mode:
    #         return None
    #
    #     raise ValueError(f"{leaf}: Not a leaf cell")

    def FilterOut(
        self,
        TrackIssues: TrackIssues_h,
        /,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        TrackIssues: Arguments are: track and (optional) keyword arguments; Returned value can be None, an str, or a
            sequence of str.
        kwargs: Passed to TrackIsInvalid as keyword arguments
        """
        t_idx = 0
        while t_idx < self.__len__():
            track = self[t_idx]
            issues = TrackIssues(track, **kwargs)
            if issues is None:
                t_idx += 1
            else:
                self.invalids.append((track, issues))
                del self[t_idx]

    def AsJsonString(self) -> str:
        """"""
        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose the contents.
        attributes = {_fld.name: getattr(self, _fld.name) for _fld in dtcl.fields(self)}

        # list(self): To hide this method, which would otherwise be infinitely recursively called
        return jsnr.JsonStringOf((list(self), attributes), true_type=self)

    def Print(self) -> None:
        """"""
        for track in self:
            print(track)

    def __repr__(self) -> str:
        """"""
        return Identity(self)

    def __str__(self) -> str:
        """"""
        return f"{self.__repr__()}:\n    {self.__len__()=}\n    {self.invalids=}\n    {self.fully_pruned=}"

    def __eq__(self, other) -> bool:
        """"""
        if hasattr(other, "AsJsonString"):
            return self.AsJsonString() == other.AsJsonString()

        return False


def UniqueCellsFromPerSingleTrackOnes(
    per_single_track: per_single_track_cells_h,
    /,
    *,
    should_remove_time_point: bool = False,
) -> cells_w_optional_time_points_h:
    """
    Per single track dividing cells => Single-track-label-indexed dictionary with cells repeated among the values.
    This function removes the redundancy among the dividing cells dictionary values.
    """
    cells = ittl.chain.from_iterable(per_single_track.values())
    if should_remove_time_point:
        cells = map(ItemAt(0), cells)

    return tuple(set(cells))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import time
from pathlib import Path as path_t
from typing import Union

import matplotlib.pyplot as pypl
import numpy as nmpy
import skimage.transform as trfm
import tifffile as tiff

from cell_tracking_BC.in_out.graphics.dbe.matplotlib.style import CellAnnotationStyle
from cell_tracking_BC.type.sequence import BoundingBoxSlices, sequence_t


def SaveSequenceAnnotations(
    sequence: sequence_t, folder: path_t, file: Union[str, path_t], /
) -> None:
    """"""
    figure, axes = pypl.subplots(facecolor="k")
    canvas = figure.canvas

    annotated = None
    for f_idx in range(sequence.length):
        if f_idx % 5 == 0:
            print(f"    Writing frame {f_idx} @ {time.ctime()}...")

        axes.clear()
        axes.set_axis_off()
        axes.set_facecolor("k")

        frame = sequence.cell_frames[f_idx]
        description = frame.Description(margins=(0, -50))
        # for contours in description["cell_contours"].values():
        #     axes.scatter(
        #         contours[1],
        #         contours[0],
        #         color=(0.0, 0.8, 0.8, 0.3),
        #     )
        for contours in description["cell_contours"].values():
            for contour in contours:
                axes.plot(
                    contour[:, 1],
                    contour[:, 0],
                    linestyle=":",
                    color=(0.0, 0.8, 0.8, 0.3),
                )

        description = sequence.DescriptionOfFrame(f_idx)
        for text, position in description["track_labels"]:
            additionals = CellAnnotationStyle(False, "\n" in text)
            axes.annotate(
                text,
                nmpy.flipud(position),
                ha="center",
                **additionals,
            )
        canvas.draw()

        content = nmpy.array(canvas.renderer.buffer_rgba())[:, :, :3]
        if annotated is None:
            annotated = nmpy.empty((*content.shape, sequence.length), dtype=nmpy.uint8)
        annotated[..., f_idx] = content

    row_slice, col_slice = BoundingBoxSlices(annotated)
    annotated = annotated[row_slice, col_slice, :, :]
    annotated = trfm.resize(
        annotated, (*sequence.shape, 3, sequence.length), preserve_range=True
    )
    annotated = annotated.astype(nmpy.uint8, copy=False)
    annotated = nmpy.moveaxis(annotated, (0, 1, 2, 3), (2, 3, 1, 0))
    annotated = annotated[:, nmpy.newaxis, :, :, :]

    tiff.imwrite(
        str(folder / file),
        annotated,
        photometric="rgb",
        compression="deflate",
        planarconfig="separate",
        metadata={"axes": "XYZCT"},
    )

    pypl.close(fig=figure)  # To prevent remaining caught in event loop



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from numbers import Real as number_t
from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import numpy as nmpy

from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.text.logger import LOGGER


class rectangle_t(list):
    edge_color: Tuple[float, float, float, float] = None
    face_color: Tuple[float, float, float, float] = None

    @classmethod
    def NewFromDetails(
        cls, corner: Tuple[int, float], width: float, height: float, /
    ) -> rectangle_t:
        """"""
        return cls((corner, width, height))

    @property
    def corner(self) -> Tuple[int, float]:
        """"""
        return self[0]

    def Details(self) -> tuple:
        """
        corner, width, height, edge_color, face_color
        """
        return *tuple(self), self.edge_color, self.face_color

    def SetHeight(self, height: float, /) -> None:
        """"""
        self[2] = height


rectangle_info_h = Tuple[int, float, float, bool, bool]
ValueTransform_h = Optional[Union[float, Callable[[float], float]]]


DIVISION_MARKER = "3"
DEATH_MARKER = "x"
DIVISION_MARKER_SIZE = 60
DEATH_MARKER_SIZE = 30

_RECTANGLE_TYPE_DIV = 0
_RECTANGLE_TYPE_DEA = 1
_RECTANGLE_TYPE_MIXED = 2


def ShowDivisionAndDeathEventResponses(
    cell_division_response: Dict[int, Optional[Sequence[float]]],
    cell_death_response: Dict[int, Optional[Sequence[float]]],
    cell_division_frame_idc: Dict[int, Optional[Sequence[int]]],
    cell_death_frame_idc: Dict[int, Optional[int]],
    sequence_length: int,
    /,
    *,
    ValueTransform: ValueTransform_h = None,
    zero_is_black: bool = True,
    show_as_barplot: bool = False,
    PlotAndShow: Callable[..., Optional[figure_t]],
    figure_name: str = "division-and-death-responses",
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    archiver: archiver_t = None,
) -> Optional[figure_t]:
    """
    Note: Does not respect pruning
    """
    max_label, div_values, dea_values, rectangles_w_details = _ValuesAndRectangles(
        cell_division_response,
        cell_death_response,
        cell_division_frame_idc,
        cell_death_frame_idc,
        show_as_barplot,
    )
    if rectangles_w_details.__len__() == 0:
        LOGGER.info("Empty division or death responses")
        return

    _SetRectanglesColorAndHeight(
        rectangles_w_details,
        div_values,
        dea_values,
        ValueTransform,
        zero_is_black,
        show_as_barplot,
    )
    events = _DivisionAndDeathEvents(rectangles_w_details)
    rectangles = tuple(_elm[0] for _elm in rectangles_w_details)

    if show_as_barplot:
        track_height = 2
    else:
        track_height = 1

    return PlotAndShow(
        events,
        rectangles,
        track_height,
        sequence_length,
        max_label,
        zero_is_black=zero_is_black,
        show_as_barplot=show_as_barplot,
        figure_name=figure_name,
        prepare_only=prepare_only,
        interactively=interactively,
        in_main_thread=in_main_thread,
        archiver=archiver,
    )


def _ValuesAndRectangles(
    cell_division_response: Dict[int, Optional[Sequence[float]]],
    cell_death_response: Dict[int, Optional[Sequence[float]]],
    cell_division_frame_idc: Dict[int, Optional[Sequence[int]]],
    cell_death_frame_idc: Dict[int, Optional[int]],
    show_as_barplot: bool,
    /,
) -> Tuple[
    int,
    Sequence[float],
    Sequence[float],
    Sequence[Tuple[rectangle_t, rectangle_info_h]],
]:
    """"""
    if (cell_division_response.__len__() == 0) or (cell_death_response.__len__() == 0):
        LOGGER.warn("Event Responses: No associated tracks.")
        return 0, (), (), ()

    max_label = max(max(cell_division_response.keys()), max(cell_death_response.keys()))
    div_values = []
    dea_values = []
    rectangles = []

    invalid_labels = set(range(1, max_label + 1))
    for label, division_response in cell_division_response.items():
        if label in cell_death_response:
            death_response = cell_death_response[label]
        else:
            death_response = None
        if label in cell_division_frame_idc:
            division_frame_idc = cell_division_frame_idc[label]
        else:
            division_frame_idc = None
        if label in cell_death_frame_idc:
            death_frame_idx = cell_death_frame_idc[label]
        else:
            death_frame_idx = None
        invalid_labels.remove(label)

        if (division_response is None) and (death_response is None):
            LOGGER.warn(f"Track {label}: No event responses")
            continue

        if division_response is None:
            LOGGER.warn(
                f"Track {label}: No division response (too short or fully pruned after death detection)"
            )
        if death_response is None:
            LOGGER.warn(
                f"Track {label}: No death response (too short or fully pruned after splitting invalidation)"
            )

        # Whether pattern-based or track-based
        death_occurred = death_frame_idx is not None
        if death_occurred:
            n_valids = abs(death_frame_idx) + 1
            if division_response is not None:
                division_response = division_response[:n_valids]
            # The death response can be None if death is geometric (negative death_frame_idx)
            if death_response is not None:
                death_response = death_response[:n_valids]

        if division_response is not None:
            div_values.extend(division_response)
        if death_response is not None:
            dea_values.extend(death_response)

        # Could be inside an "if not show_as_barplot", but the IDE might complain about potentially uninitialized object
        rtype = _RECTANGLE_TYPE_MIXED
        if division_response is None:
            if show_as_barplot:
                rtype = _RECTANGLE_TYPE_DEA
            for t_idx, dea_value in enumerate(death_response):
                if dea_value is None:
                    # The trajectory does not start at time point 0
                    continue

                # Note: division_frame_idc is None if no divisions (used to be (-1,))
                # Note: division is checked despite the track having no division response because the sibling track
                # might have a response (if long enough) higher enough.
                dividing = (division_frame_idc is not None) and (
                    t_idx in division_frame_idc
                )
                # Detected by pattern matching (track ending => negative death frame index)
                dying = death_occurred and (t_idx == death_frame_idx)

                rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                details = (rtype, 0.0, dea_value, dividing, dying)
                rectangles.append((rectangle, details))

                if dying:
                    break
        elif death_response is None:
            if show_as_barplot:
                rtype = _RECTANGLE_TYPE_DIV
            for t_idx, div_value in enumerate(division_response):
                if div_value is None:
                    # The trajectory does not start at time point 0
                    continue

                # Note: division_frame_idc is None if no divisions (used to be (-1,))
                dividing = (division_frame_idc is not None) and (
                    t_idx in division_frame_idc
                )

                rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                details = (rtype, div_value, 0.0, dividing, False)
                rectangles.append((rectangle, details))
        else:
            for t_idx, (div_value, dea_value) in enumerate(
                zip(division_response, death_response)
            ):
                if div_value is None:
                    # The trajectory does not start at time point 0
                    if dea_value is not None:
                        raise ValueError(
                            f"{label}: Trajectory has a valid death response at time point {t_idx} "
                            f"whereas the division response is invalid"
                        )
                    continue

                # Note: division_frame_idc is None if no divisions (used to be (-1,))
                dividing = (division_frame_idc is not None) and (
                    t_idx in division_frame_idc
                )
                # Detected by pattern matching (track ending => negative death frame index)
                dying = death_occurred and (t_idx == death_frame_idx)

                rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                if show_as_barplot:
                    details = (_RECTANGLE_TYPE_DEA, 0.0, dea_value, False, dying)
                    rectangles.append((rectangle, details))

                    rectangle = _EventRectangle(t_idx, label, show_as_barplot)
                    details = (_RECTANGLE_TYPE_DIV, div_value, 0.0, dividing, False)
                else:
                    details = (
                        _RECTANGLE_TYPE_MIXED,
                        div_value,
                        dea_value,
                        dividing,
                        dying,
                    )
                rectangles.append((rectangle, details))

                if dying:
                    break

    LOGGER.warn(f"Invalid Tracks (no event responses): {invalid_labels}")

    return max_label, div_values, dea_values, rectangles


def _EventRectangle(
    time_point: int,
    label: int,
    show_as_barplot: bool,
    /,
) -> rectangle_t:
    """"""
    if show_as_barplot:
        corner_y = 2.0 * label - 1.5
    else:
        corner_y = label - 0.5

    return rectangle_t.NewFromDetails((time_point, corner_y), 1.0, 1.0)


def _SetRectanglesColorAndHeight(
    rectangles: Sequence[Tuple[rectangle_t, rectangle_info_h]],
    div_values: Sequence[float],
    dea_values: Sequence[float],
    ValueTransform: ValueTransform_h,
    zero_is_black: bool,
    show_as_barplot: bool,
    /,
) -> None:
    """
    ValueTransform: if a number, then value <= ValueTransform => value <- 0.0
    """
    if ValueTransform is None:
        ValueTransform = lambda _vle: _vle
    elif isinstance(ValueTransform, number_t):
        threshold = float(ValueTransform)

        def _ValueTransform(_vle: float, /) -> float:
            if _vle <= threshold:
                return 0.0
            else:
                return _vle

        ValueTransform = _ValueTransform

    min_div_value, div_scaling = _MinimumAndScaling(div_values)
    min_dea_value, dea_scaling = _MinimumAndScaling(dea_values)

    for rectangle, (rtype, div_value, dea_value, dividing, dying) in rectangles:
        div = ValueTransform(div_scaling * (div_value - min_div_value))
        dea = ValueTransform(dea_scaling * (dea_value - min_dea_value))

        if show_as_barplot:
            if rtype == _RECTANGLE_TYPE_DEA:
                if zero_is_black:
                    color = (dea, 0.0, 0.0)
                else:
                    color = (1.0, 1.0 - dea, 1.0 - dea)
                rectangle.SetHeight(1.5 * dea)
            else:
                if zero_is_black:
                    color = (0.0, 0.0, div)
                else:
                    color = (1.0 - div, 1.0 - div, 1.0)
                rectangle.SetHeight(1.5 * div)
                rectangle.face_color = color + (0.3,)
            rectangle.edge_color = color
        else:
            if zero_is_black:
                color = (dea, 0.0, div)
            else:
                div_color = (1.0 - div, 1.0 - div, 1.0)
                dea_color = (1.0, 1.0 - dea, 1.0 - dea)
                color = nmpy.minimum(div_color, dea_color)
            rectangle.edge_color = color
            rectangle.face_color = color


def _DivisionAndDeathEvents(
    rectangles: Sequence[Tuple[rectangle_t, rectangle_info_h]],
    /,
) -> Sequence[Tuple[str, int, Sequence[float]]]:
    """"""
    output = []

    for rectangle, (*_, dividing, dying) in rectangles:
        if dividing or dying:
            where = tuple(_crd + 0.5 for _crd in rectangle.corner)
            if dividing:
                output.append((DIVISION_MARKER, DIVISION_MARKER_SIZE, where))
            if dying:
                output.append((DEATH_MARKER, DEATH_MARKER_SIZE, where))

    return output


def _MinimumAndScaling(values: Sequence[float], /) -> Tuple[float, float]:
    """"""
    if values.__len__() > 0:
        min_value = min(values)
        max_value = max(values)
        scaling = 1.0 / (max_value - min_value)
    else:
        min_value, scaling = 0.0, 1.0

    return min_value, scaling



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from sys import maxsize as MAX_INTEGER
from typing import Any, List, Optional, Sequence, Tuple

import networkx as grph
import numpy as nmpy

from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t as axes_t
from cell_tracking_BC.in_out.graphics.type.color import colormap_h, rgba_color_h
from cell_tracking_BC.in_out.graphics.type.context import context_t, path_collection_h
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.track import VersionOfForkingForLayout
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.type.track.forking import forking_track_t
from cell_tracking_BC.type.track.single import single_track_t
from cell_tracking_BC.type.tracks.structured import tracks_t


_EDGE_COLORS = {True: "gray", False: "red"}
_TICK_COLORS = {True: "black", False: "red"}


# TODO: factorization and simplification needed


@dtcl.dataclass(repr=False, eq=False)
class t_viewer_2d_t:

    figure: figure_t
    axes: axes_t
    colormap: colormap_h

    scatter: path_collection_h = dtcl.field(init=False, default=None)

    # Cell details
    labels: List[int] = dtcl.field(init=False, default_factory=list)
    time_points: List[int] = dtcl.field(init=False, default_factory=list)
    affinities: List[float] = dtcl.field(init=False, default_factory=list)
    colors: List[rgba_color_h] = dtcl.field(init=False, default_factory=list)

    dbe: context_t = None

    @classmethod
    def NewForTracks(
        cls,
        tracks: tracks_t,
        dbe: context_t,
        /,
        *,
        mode: str = "forking",
        in_axes: axes_t = None,
    ) -> t_viewer_2d_t:
        """
        mode: single or forking
        """
        if mode not in ("single", "forking"):
            raise ValueError(
                f'{mode}: Invalid plotting mode. Expected="single" or "forking".'
            )

        if in_axes is None:
            figure, axes = dbe.figure_2d_t.NewFigureAndAxes()
        else:
            figure = in_axes.Figure()
            axes = in_axes
        axes.SetTrackingAxesProperties(
            range(1, sum(_tck.n_leaves for _tck in tracks) + 1)
        )
        colormap = axes.AddStandardColormap(
            "Tracking Affinity", "plasma", position="left"
        )

        instance = cls(
            figure=figure,
            axes=axes,
            colormap=colormap,
            dbe=dbe,
        )

        all_cell_heights = []
        all_tracks = list(tracks)
        limit_valid = all_tracks.__len__()
        all_tracks.extend(_rcd[0] for _rcd in tracks.invalids if isinstance(_rcd[0], (single_track_t, forking_track_t)))
        limit_invalid = all_tracks.__len__()
        all_tracks.extend(tracks.fully_pruned)
        tick_details = []
        for t_idx, track in enumerate(all_tracks):
            is_valid = t_idx < limit_valid
            is_fully_pruned = t_idx >= limit_invalid
            if is_fully_pruned:
                min_label = min(track.pruned_labels)
                n_geometric_leaves = track.geometric_leaves.__len__()
                heights = tuple(range(min_label, min_label + n_geometric_leaves))
                tick_details.extend((_hgt, "fp", "magenta") for _hgt in heights)
                continue

            if isinstance(track, single_track_t):
                PlotTrackEdges = instance._PlotSingleTrackEdges
            elif mode == "single":
                PlotTrackEdges = instance._PlotForkingTrackEdgesAsSingle
            else:
                PlotTrackEdges = instance._PlotForkingTrackEdges
            new_labels = PlotTrackEdges(track, all_cell_heights, is_valid)
            tick_details.extend(new_labels)

        instance.scatter = axes.PlotPoints(
            instance.time_points,
            all_cell_heights,
            marker="o",
            c=instance.colors,
            zorder=2,
        )
        positions, labels, colors = zip(*tick_details)
        axes.SetAxisTicks("yaxis", positions, labels, colors=colors)

        return instance

    def _PlotSingleTrackEdges(
        self,
        track: single_track_t,
        all_cell_heights: List[int],
        is_valid: bool,
        /,
    ) -> Sequence[Tuple[float, int, Any]]:
        """"""
        length, time_points, label, where = self._ElementsForTrackPiece(
            track.unpruned_piece
        )

        self.labels.insert(where, track.root.label)
        self.time_points.insert(where, track.root_time_point)
        self.affinities.insert(where, 0.0)
        self.colors.insert(where, self.colormap(0.0))

        heights = (length + 1) * (label,)
        all_cell_heights.extend(heights)

        self.axes.PlotLines(
            time_points, heights, color=_EDGE_COLORS[is_valid], zorder=1
        )

        return ((label, label, _TICK_COLORS[is_valid]),)  # Note: this is a tuple

    def _PlotForkingTrackEdgesAsSingle(
        self,
        track: forking_track_t,
        all_cell_heights: List[int],
        is_valid: bool,
        /,
    ) -> Sequence[Tuple[float, int, Any]]:
        """"""
        output = []

        for single in track.single_tracks_iterator:
            height_label = self._PlotSingleTrackEdges(
                single, all_cell_heights, is_valid
            )
            output.extend(height_label)

        return output

    def _PlotForkingTrackEdges(
        self,
        track: forking_track_t,
        all_cell_heights: List[int],
        is_valid: bool,
        /,
    ) -> Sequence[Tuple[float, int, Any]]:
        """"""
        with_int_labels, integer_to_cell = VersionOfForkingForLayout(track)
        try:
            int_layout = grph.nx_agraph.pygraphviz_layout(with_int_labels, prog="dot")
        except Exception as exc:
            LOGGER.warn(
                f"Track layout failed for track {track.labels} with error:\n{exc}"
            )
            return ()
        positions = {
            integer_to_cell[_idx]: _pst
            for _idx, _pst in int_layout.items()
            if isinstance(_idx, int)
        }

        output = []

        all_time_points, all_heights = [], []
        min_height = max_height = positions[track.root][0]
        min_label = MAX_INTEGER
        root_height = None
        where = None
        for piece in track.Pieces():
            _, time_points, label, new_where = self._ElementsForTrackPiece(piece)
            heights = nmpy.fromiter(
                (positions[_cll][0] for _cll in piece), dtype=nmpy.float64
            )
            if piece[0] is track.root:
                root_height = heights[0]
            if where is None:
                where = new_where

            all_time_points.append(time_points)
            all_heights.append(heights)
            min_height = min(min_height, min(heights))
            max_height = max(max_height, max(heights))
            if label is not None:
                if label < min_label:
                    min_label = label
                output.append((heights[-1], label, _TICK_COLORS[is_valid]))

        # max(max_height - min_height, 1): Because max_height - min_height can be zero if the track has been pruned so
        # that it now has the structure of a single track.
        height_scaling = (track.n_leaves - 1) / max(max_height - min_height, 1)
        AdjustedHeight = lambda _hgt: height_scaling * (_hgt - min_height) + min_label

        output = tuple(
            (AdjustedHeight(_elm[0]), _elm[1], _TICK_COLORS[is_valid])
            for _elm in sorted(output)
        )

        output += tuple((_hgt, "p", "magenta") for _hgt in track.pruned_labels)

        for time_points, heights in zip(all_time_points, all_heights):
            heights = AdjustedHeight(heights)
            self.axes.PlotLines(
                time_points, heights, color=_EDGE_COLORS[is_valid], zorder=1
            )
            all_cell_heights.extend(heights[1:])
        root_height = AdjustedHeight(root_height)

        self.labels.insert(where, track.root.label)
        self.time_points.insert(where, track.root_time_point)
        self.affinities.insert(where, 0.0)
        self.colors.insert(where, self.colormap(0.0))

        all_cell_heights.insert(where, root_height)

        return output

    def _ElementsForTrackPiece(
        self,
        piece: single_track_t,
        /,
    ) -> Tuple[int, Sequence[int], Optional[int], int]:
        """"""
        where = self.labels.__len__()

        label = piece.label
        root_time_point = piece.root_time_point
        length = piece.length
        time_points = tuple(range(root_time_point, root_time_point + length + 1))
        affinities = piece.affinities

        self.labels.extend(_cll.label for _cll in piece[1:])
        self.time_points.extend(time_points[1:])
        self.affinities.extend(affinities)
        self.colors.extend(self.colormap(affinities))

        return length, time_points, label, where



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import difflib as diff
from typing import Any

import numpy as nmpy

from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.compartment import compartment_t
from cell_tracking_BC.type.frame import frame_t


array_t = nmpy.ndarray


def CompareInstances(element_1: Any, element_2: Any, /) -> int:
    """"""
    if element_1 == element_2:
        return 0
    else:
        if hasattr(element_1, "AsJsonString"):
            as_str = element_1.AsJsonString()
            as_str_1 = tuple(
                as_str[c_idx : (c_idx + 300)]
                for c_idx in range(0, as_str.__len__(), 300)
            )

            as_str = element_2.AsJsonString()
            as_str_2 = tuple(
                as_str[c_idx : (c_idx + 300)]
                for c_idx in range(0, as_str.__len__(), 300)
            )

            differences = "\n".join(
                diff.context_diff(as_str_1, as_str_2, fromfile="before", tofile="after")
            )
            print(f"{type(element_1).__name__}: Mismatch:\n{differences}")
        else:
            print(
                f"{type(element_1).__name__}: Mismatch:\n{element_1}\n!=\n{element_2}"
            )
        return 1


if __name__ == "__main__":
    #
    map_ = nmpy.zeros((10, 10))
    map_[4:7, 4:7] = 1

    for original in (
        compartment_t.NewFromMap(map_),
        cell_t.NewFromMaps(1, cell_map=map_),
        frame_t(map_),
    ):
        jsoned = original.AsJsonString()
        decoded = type(original).NewFromJsonString(jsoned)
        CompareInstances(original, decoded)



# from typing import Tuple, Union

import numpy as nmpy

import scipy.interpolate as trpl
import skimage as skim


array_t = nmpy.ndarray


_LOCAL_HISTOGRAM_FILTER_WIDTH = 3.0
_RESCALING_FILTER_WIDTH = 9.0


def WithBackgroundSubtracted(
    image: array_t,
    segmentation: array_t,
    /,
    *,
    mode: str = None,
    dilation: int = 5,
) -> array_t:
    """
    mode: None (=mean), mean, median, interpolated
    """
    missing_map = skim.morphology.binary_dilation(
        segmentation, selem=skim.morphology.disk(dilation)
    )
    filled_map = nmpy.logical_not(missing_map)

    if (mode is None) or (mode == "mean"):
        output = image - nmpy.mean(image[filled_map])
    elif mode == "median":
        output = image - nmpy.median(image[filled_map])
    elif mode == "interpolated":
        output = image - _PChipInterpolatedBackground(image, filled_map, missing_map)
    else:
        raise NotImplementedError(
            f"{mode}: Invalid or unimplemented background subtraction mode"
        )

    output[output < 0.0] = 0.0

    # import matplotlib.pyplot as p
    # p.matshow(image), p.colorbar()
    # a=image.copy()
    # a[missing_map] = 0
    # p.matshow(a), p.colorbar()
    # p.matshow(_PChipInterpolatedBackground(image, filled_map, missing_map)), p.colorbar()
    # p.matshow(output), p.colorbar()
    # p.show()

    return output


def _PChipInterpolatedBackground(
    incomplete: array_t, filled_map: array_t, missing_map: array_t, /
) -> array_t:
    """
    Uses pchip interpolation to avoid negative values
    """
    # Since pchip works in 1-D, interpolation has to be done in 2 steps
    # Step 1: row-wise interpolation
    row_wise = nmpy.array(incomplete, dtype=nmpy.float64)

    for col in range(row_wise.shape[1]):
        old_rows = nmpy.nonzero(filled_map[:, col])[0]
        new_rows = nmpy.nonzero(missing_map[:, col])[0]

        row_wise[new_rows, col] = trpl.pchip_interpolate(
            old_rows, incomplete[old_rows, col], new_rows
        )

    # Step 2: column-wise interpolation
    col_wise = nmpy.array(incomplete, dtype=nmpy.float64)

    for row in range(col_wise.shape[0]):
        old_cols = nmpy.nonzero(filled_map[row, :])[0]
        new_cols = nmpy.nonzero(missing_map[row, :])[0]

        col_wise[row, new_cols] = trpl.pchip_interpolate(
            old_cols, incomplete[row, old_cols], new_cols
        )

    output = nmpy.array(incomplete, dtype=nmpy.float64)
    output[missing_map] = 0.5 * (row_wise[missing_map] + col_wise[missing_map])

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from typing import Callable, Iterator, Optional, Sequence, Tuple

import networkx as grph

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.standard.issue import ISSUE_SEPARATOR
from cell_tracking_BC.type.cell import cell_t


# TODO: check if grph.descendants and grph.ancestors could be used in place of more complex code here
# TODO: cell iteration methods should have a geometric_mode: bool = False parameter to allow cell-state-independent
#     iterations


TIME_POINT = "time_point"  # Leave it here (as opposed to making it a class variable) since it is used "everywhere"


class unstructured_track_t(grph.DiGraph):
    in_degree: Callable[[cell_t], int]
    out_degree: Callable[[cell_t], int]

    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> unstructured_track_t:
        """"""
        return cls(
            jsnr.ObjectFromJsonString(
                jsoned, builders={"cell_t": cell_t.NewFromJsonString}
            )
        )

    def Issues(self) -> Optional[Sequence[str]]:
        """"""
        output = []

        for cell in self.nodes:
            if (n_predecessors := self.in_degree(cell)) > 1:
                output.append(
                    f"Cell {cell.label}{ISSUE_SEPARATOR}{n_predecessors} predecessors. Expected=0 or 1."
                )

        if output.__len__() == 0:
            output = None

        return output

    def RootCellWithTimePoint(self) -> Tuple[cell_t, int]:
        """"""
        output = tuple(
            _rcd for _rcd in self.nodes.data(TIME_POINT) if self.in_degree(_rcd[0]) == 0
        )

        if (n_roots := output.__len__()) != 1:
            raise ValueError(f"{n_roots}: Invalid number of root cells. Expected=1.")

        return output[0]

    def LeafCellsWithTimePoints(self) -> Tuple[Sequence[cell_t], Sequence[int]]:
        """"""
        # TODO: Contact the Networkx team about the following comment (or check code on github)
        #     /!\ It seems that networkx.DiGraph.nodes.data does not guarantee the node enumeration order. This could be
        #     inconvenient for reproducibility checks.
        records = (
            _rcd
            for _rcd in self.nodes.data(TIME_POINT)
            if self.out_degree(_rcd[0]) == 0
        )
        leaves, time_points = zip(*records)

        return tuple(leaves), tuple(time_points)

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        time_points = grph.get_node_attributes(self, TIME_POINT)

        for edge in self.edges:
            time_point = time_points[edge[0]]
            is_last = self.out_degree(edge[1]) == 0
            yield time_point, *edge, is_last

    def AsJsonString(self) -> str:
        """"""
        return jsnr.JsonStringOf(self)

    def __eq__(self, other) -> bool:
        """"""
        if hasattr(other, "AsJsonString"):
            return self.AsJsonString() == other.AsJsonString()

        return False
